"""
The code is borrowed from:
-------------------------------------------------------------------------------
https://github.com/anilbatra2185/road_connectivity/blob/master/model/linknet.py
"""

import math
from typing import Tuple

import numpy as np
import torch
#from arcgis.learn._utils.segmentation_loss_functions import mIoULoss
from fastai.callbacks.hooks import hook_outputs, model_sizes
from torch import nn


class DecoderBlock(nn.Module):
    """
    Decoder Block to perform Transposed Convolutions
    @param in_channels: Number of input channels
    @param out_channels: Number of output channels
    @param group: grouping of channels used in convolution
                Defaut is set to 1 i.e. vanilla convolution
    """

    def __init__(self, in_channels, out_channels, hook=None, group=1):
        super(DecoderBlock, self).__init__()

        self.hook = hook
        # B, C, H, W -> B, C/4, H, W
        self.conv1 = nn.Conv2d(in_channels, in_channels // 4, 1, groups=group)
        self.norm1 = nn.BatchNorm2d(in_channels // 4)
        self.relu1 = nn.ReLU(inplace=True)

        # B, C/4, H, W -> B, C/4, H, W
        self.deconv2 = nn.ConvTranspose2d(
            in_channels // 4,
            in_channels // 4,
            3,
            stride=2,
            padding=1,
            output_padding=1,
            groups=group,
        )
        self.norm2 = nn.BatchNorm2d(in_channels // 4)
        self.relu2 = nn.ReLU(inplace=True)

        # B, C/4, H, W -> B, C, H, W
        self.conv3 = nn.Conv2d(in_channels // 4, out_channels, 1, groups=group)
        self.norm3 = nn.BatchNorm2d(out_channels)
        self.relu3 = nn.ReLU(inplace=True)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
            if isinstance(m, nn.ConvTranspose2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def forward(self, x):
        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)
        x = self.deconv2(x)
        x = self.norm2(x)
        x = self.relu2(x)
        x = self.conv3(x)
        x = self.norm3(x)
        x = self.relu3(x)
        if self.hook:
            x = x + self.hook.stored
        return x


class LinkNetModel(nn.Module):
    """
    Implements LinkNet model.
    Reference: https://arxiv.org/pdf/1707.03718.pdf
    """
    def __init__(self, encoder, n_classes, chip_size, n_bands=3):
        super(LinkNetModel, self).__init__()

        self.chip_size = chip_size
        # Fetch the sizes of various activation layers of the backbone
        sfs_sizes = model_sizes(encoder, size=self.chip_size)
        filters = [x[1] for x in sfs_sizes[-4:]]

        # Encoder
        self.encoder = encoder
        self.encoder_outputs = hook_outputs(self.encoder)

        # Decoder
        self.decoder4 = DecoderBlock(filters[3], filters[2], self.encoder_outputs[-2])
        self.decoder3 = DecoderBlock(filters[2], filters[1], self.encoder_outputs[-3])
        self.decoder2 = DecoderBlock(filters[1], filters[0], self.encoder_outputs[-4])
        self.decoder1 = DecoderBlock(filters[0], filters[0])

        # Final Classifier
        self.finaldeconv1 = nn.ConvTranspose2d(filters[0], 32, 3, stride=2)
        self.finalrelu1 = nn.LeakyReLU(0.2, inplace=True)
        self.finalconv2 = nn.Conv2d(32, 32, 3)
        self.finalrelu2 = nn.LeakyReLU(0.2, inplace=True)
        self.finalconv3 = nn.Conv2d(32, n_classes, 2, padding=1)
        self.sigmoid = nn.Sigmoid()

        for m in [self.finaldeconv1, self.finalconv2]:
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
            if isinstance(m, nn.ConvTranspose2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def forward(self, x):
        rows = x.size()[2]
        cols = x.size()[3]

        e = self.encoder(x)
        d4 = self.decoder4(e)
        d3 = self.decoder3(d4)
        d2 = self.decoder2(d3)
        d1 = self.decoder1(d2)

        # Final Classification
        f1 = self.finaldeconv1(d1)
        f2 = self.finalrelu1(f1)
        f3 = self.finalconv2(f2)
        f4 = self.finalrelu2(f3)
        f5 = self.finalconv3(f4)
        return f5[:, :, :rows, :cols]


class LinkNetMultiTaskModel(nn.Module):
    """
    Implements LinkNet MultiTask model.
    Reference: https://arxiv.org/pdf/1707.03718.pdf
    """

    def __init__(self, encoder, task1_classes, task2_classes, chip_size, n_bands=3):
        super(LinkNetMultiTaskModel, self).__init__()

        self.chip_size = chip_size
        # Fetch the sizes of various activation layers of the backbone
        sfs_sizes = model_sizes(encoder, size=self.chip_size)
        filters = [x[1] for x in sfs_sizes[-4:]]

        # Encoder
        self.encoder = encoder
        self.encoder_outputs = hook_outputs(self.encoder)

        # Decoder for Road Segmentation
        self.decoder4 = DecoderBlock(filters[3], filters[2], self.encoder_outputs[-2])
        self.decoder3 = DecoderBlock(filters[2], filters[1], self.encoder_outputs[-3])
        self.decoder2 = DecoderBlock(filters[1], filters[0], self.encoder_outputs[-4])
        self.decoder1 = DecoderBlock(filters[0], filters[0])

        # Final Classifier for Road Segmentation
        self.finaldeconv1 = nn.ConvTranspose2d(filters[0], 32, 3, stride=2)
        self.finalrelu1 = nn.LeakyReLU(0.2, inplace=True)
        self.finalconv2 = nn.Conv2d(32, 32, 3)
        self.finalrelu2 = nn.LeakyReLU(0.2, inplace=True)
        self.finalconv3 = nn.Conv2d(32, task1_classes, 2, padding=1)

        # Decoder for Road Orientation
        self.o_decoder4 = DecoderBlock(filters[3], filters[2], self.encoder_outputs[-2])
        self.o_decoder3 = DecoderBlock(filters[2], filters[1], self.encoder_outputs[-3])
        self.o_decoder2 = DecoderBlock(filters[1], filters[0], self.encoder_outputs[-4])
        self.o_decoder1 = DecoderBlock(filters[0], filters[0])

        # Final Classifier for Road Orientation
        self.o_finaldeconv1 = nn.ConvTranspose2d(filters[0], 32, 3, stride=2)
        self.o_finalrelu1 = nn.LeakyReLU(0.2, inplace=True)
        self.o_finalconv2 = nn.Conv2d(32, 32, 3)
        self.o_finalrelu2 = nn.LeakyReLU(0.2, inplace=True)
        self.o_finalconv3 = nn.Conv2d(32, task2_classes, 2, padding=1)

        for m in [self.finaldeconv1, self.finalconv2, self.o_finaldeconv1, self.o_finalconv2]:
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
            if isinstance(m, nn.ConvTranspose2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def forward(self, x):
        rows = x.size()[2]
        cols = x.size()[3]

        e = self.encoder(x)

        # Decoder - Road Segmentation
        d4 = self.decoder4(e)
        d3 = self.decoder3(d4)
        d2 = self.decoder2(d3)
        d1 = self.decoder1(d2)

        # Final Classification - Road Segmentation
        f1 = self.finaldeconv1(d1)
        f2 = self.finalrelu1(f1)
        f3 = self.finalconv2(f2)
        f4 = self.finalrelu2(f3)
        f5 = self.finalconv3(f4)

        # Decoder - Road Orientation
        o_d4 = self.o_decoder4(e)
        o_d3 = self.o_decoder3(o_d4)
        o_d2 = self.o_decoder2(o_d3)
        o_d1 = self.o_decoder1(o_d2)

        # Final Classification - Road Segmentation
        o_f1 = self.o_finaldeconv1(o_d1)
        o_f2 = self.o_finalrelu1(o_f1)
        o_f3 = self.o_finalconv2(o_f2)
        o_f4 = self.o_finalrelu2(o_f3)
        o_f5 = self.o_finalconv3(o_f4)

        return f5[:, :, :rows, :cols], o_f5[:, :, :rows, :cols]


class road_orient_loss(nn.Module):
    """
    Custom PyTorch Loss Module to compute loss for backpropagation of multi-task
    outputs. It compose of Mean-IoU and CrossEntropy losses.
    In current version it only supports Mean-IoU for Road Segmentation and
    Cross-Entropy for Road Orientations.
    """
    def __init__(self, n_classes: int = 2, loss_weights: Tuple = (1.0, 1.0)):
        super().__init__()
        self.n_classes = n_classes
        self.loss_weights = loss_weights if loss_weights else (1.0, 1.0)
        self.road_loss = nn.CrossEntropyLoss()  # mIoULoss(self.n_classes)
        self.orient_loss = nn.CrossEntropyLoss()

    def forward(self, predictions, *target):
        if isinstance(predictions, (list, tuple)):
            if len(target) > 1:
                seg_target, orient_target = target
                return self.loss_weights[0] * self.road_loss(predictions[0], seg_target.squeeze(1).long()) + \
                       self.loss_weights[1] * self.orient_loss(predictions[1], orient_target.squeeze(1).long())
            else:
                seg_target = target[0]
                return self.loss_weights[0] * self.road_loss(predictions[0], seg_target.squeeze(1).long())
        else:
            return self.road_loss(predictions, target[0].squeeze(1).long())


def isin(target, keep_indices):
    old_shape = target.shape
    mask = torch.cat(
        [(target.view(-1) == k)[:, None] for k in keep_indices], dim=1
    ).any(1)
    mask = mask.view(old_shape).contiguous()
    return mask


def accuracy(input, *target, ignore_mapped_class=[]):
    """
    Compute pixel based accuracy
    """
    if isinstance(input, tuple):  # while training
        input = input[0]
    if isinstance(target, tuple):
        target = target[0]

    target = target.squeeze(1).cpu().long()
    input = input.cpu()
    if ignore_mapped_class == []:
        return (input.argmax(dim=1) == target).float().mean()
    else:
        _, total_classes, _, _ = input.shape
        keep_indices = [i for i in range(total_classes) if i not in ignore_mapped_class]
        for k in ignore_mapped_class:
            input[:, k] = -1
        targ_mask = isin(target, keep_indices)
        return (input.argmax(dim=1)[targ_mask] == target[targ_mask]).float().mean()


def miou(prediction, *target, ignore_mapped_class=[], smooth=1e-8):
    """
    Compute pixel based Mean IoU
    """
    if isinstance(prediction, tuple):  # while training
        prediction = prediction[0]
    if isinstance(target, tuple):
        target = target[0]

    def fast_hist(a, b, n):
        k = (a >= 0) & (a < n)
        return np.bincount(n * a[k].astype(int) + b[k], minlength=n ** 2).reshape(n, n)

    target = target.squeeze(1).long()
    batch_size = prediction.size(0)
    classes = prediction.size(1)
    prediction = prediction.argmax(dim=1)

    histogram = fast_hist(
        prediction.view(batch_size, -1).cpu().numpy(),
        target.view(batch_size, -1).cpu().numpy(),
        classes,
    )
    iou_per_class = np.diag(histogram) / (
        histogram.sum(1) + histogram.sum(0) - np.diag(histogram)
    )

    if ignore_mapped_class == []:
        return torch.tensor(np.nanmean(iou_per_class))
    else:
        sum = 0.0
        mean_classes = 0
        for k in classes:
            if k not in ignore_mapped_class:
                mean_classes += 1
                sum += iou_per_class[k]
        if mean_classes == 0:
            return 0
        return torch.tensor(sum / mean_classes)



from arcgis._impl.common._isd import InsensitiveDict
###########################################################################
class APIKey(object):
    """
    The `APIKey` is a single instance of a registered access key for 
    performing certian operations based on permissions.
    """
    _gis = None
    _item = None
    _properties = None
    def __init__(self, item, gis):
        self._item = item
        self._gis = gis
    #----------------------------------------------------------------------
    def __str__(self):
        return f"<API Key {self._item.title}>"
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()    
    #----------------------------------------------------------------------
    @property
    def properties(self):
        if self._properties is None:
            self._properties = InsensitiveDict(self._item.app_info)
        return self._properties
    #----------------------------------------------------------------------
    @property
    def apikey(self):
        """
        Returns the API Key value for the current key.
        
        :returns: String
        """
        return self.properties.apiKey
    #----------------------------------------------------------------------
    def delete(self):
        return self._item.delete()
    #----------------------------------------------------------------------
    def reset(self):
        """
        Resets the API Key for the Item. The call will return the information 
        with the new API Key information.
        
        :returns: dict
        
        """
        url = f"{self._gis._portal.resturl}oauth2/apps/{self.properties.client_id}/resetApiKey"
        params = {'f' : 'json'}
        self._properties = None
        return self._gis._con.post(url, params)
    #----------------------------------------------------------------------
    def update(self, http_referers=None, privileges=None):
        """
        Updates the API Key's properties
        
        ================  ===============================================================================
        **Parameter**     **Description**
        ----------------  -------------------------------------------------------------------------------
        http_referers     Optional List. A list of the http referrers for which usage of the 
                          API Key will be restricted to.
                          
                          **Example**
                            
                          ```
                          [
                          "https://foo.com",
                          "https://bar.com"
                          ]
                          ```
                          
                          Note: Http Referrers can be configured for non apiKey type apps as 
                          well. The list configured here will be used to validate the app 
                          tokens sent in while accessing the sharing API. The referrer checks 
                          will not be applied to user tokens.
        ----------------  -------------------------------------------------------------------------------
        privileges        Optional List. A list of the privileges that will be available for 
                          this API key.
                          
                          **Example**
                          
                          ```   
                          
                          [
                          "portal:apikey:basemaps",
                          "portal:app:access:item:itemId",
                          "premium:user:geocode",
                          "premium:user:networkanalysis"
                          ]
                          
                          ```

                          Note: Privileges can be configured for non  `API Key` type apps as 
                          well. The list configured here will be used to grant access to items
                          when item endpoint is accessed with app tokens. The checks will not 
                          be applied to user tokens and they can continue accessing items 
                          based on the current item sharing model. With app tokens, all items 
                          of app owner can be accessed if the privileges list is not 
                          configured.
        ================  ===============================================================================
        
        :returns: dict
        
        """
        url = f"{ self._gis._portal.resturl}oauth2/apps/{self.properties.client_id}/update"
        if http_referers is None and privileges is None:
            return self.properties
        params = {'f' : 'json'}
        if http_referers:
            params['httpReferrers'] = http_referers
        if privileges:
            params['priveleges'] = privileges
        self._properties = None
        return self._gis._con.post(url, params)
###########################################################################
class APIKeyManager(object):
    """
    Creates, manages and updates API Keys for ArcGIS Online
    """
    _gis = None
    _url = None
    def __init__(self, gis):
        from .. import GIS
        assert isinstance(gis, GIS)
        self._gis = gis 
        self._base_url = f"{ self._gis._portal.resturl}"
    #----------------------------------------------------------------------
    def __str__(self):
        return f"<API Key Manager @ {self._gis._portal.resturl}>"
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()
    #----------------------------------------------------------------------
    def get(self, api_key):
        """
        Returns a Single API Key based on the Key Value
        """
        for key in self.keys:
            if key.properties.apikey.lower() == api_key.lower():
                return key
        return None
    #----------------------------------------------------------------------
    def create(self, 
               title, 
               tags, 
               description=None, 
               http_referers=None, 
               redirect_uris=None, 
               privileges=None):
        """
        Generates a new API Key for the Organization.
        
        ================  ===============================================================================
        **Parameter**     **Description**
        ----------------  -------------------------------------------------------------------------------
        title             Required String. The name of the API Key Item.
        ----------------  -------------------------------------------------------------------------------
        tags              Required String. A comma seperated list of descriptive words describing the 
                          API Key item.
        ----------------  -------------------------------------------------------------------------------
        description       Optional String. A description of what the API Key is going to be used for.
        ----------------  -------------------------------------------------------------------------------
        http_referers     Optional List. A list of the http referrers for which usage of the 
                          API Key will be restricted to.
                          
                          **Example**
                            
                          ```
                          [
                          "https://foo.com",
                          "https://bar.com"
                          ]
                          ```
                          
                          Note: Http Referrers can be configured for non apiKey type apps as 
                          well. The list configured here will be used to validate the app 
                          tokens sent in while accessing the sharing API. The referrer checks 
                          will not be applied to user tokens.
        ----------------  -------------------------------------------------------------------------------
        redirect_uris     Optional list.  The URIs where the access_token or authorization
                          code will be delivered upon successful authorization. The
                          redirect_uri specified during authorization must match one of the
                          registered URIs, otherwise authorization will be rejected.
                          
                          A special value of urn:ietf:wg:oauth:2.0:oob can also be specified
                          for authorization grants. This will result in the authorization
                          code being delivered to a portal URL (/oauth2/approval). This
                          value is typically used by apps that don't have a web server or a
                          custom URI scheme where the code can be delivered.
                          
                          The value is a JSON string array.
                          
                          Example:

                            [
                                "https://app.example.com",
                                "urn:ietf:wg:oauth:2.0:oob"
                            ]
                          
        
        ----------------  -------------------------------------------------------------------------------
        privileges        Optional List. A list of the privileges that will be available for 
                          this API key.
                          
                          **Example**
                          
                          ```   
                          
                          [
                          "portal:apikey:basemaps",
                          "portal:app:access:item:itemId",
                          "premium:user:geocode",
                          "premium:user:networkanalysis"
                          ]
                          
                          ```

                          Note: Privileges can be configured for non  `API Key` type apps as 
                          well. The list configured here will be used to grant access to items
                          when item endpoint is accessed with app tokens. The checks will not 
                          be applied to user tokens and they can continue accessing items 
                          based on the current item sharing model. With app tokens, all items 
                          of app owner can be accessed if the privileges list is not 
                          configured.
        ================  ===============================================================================
        
        :returns: `APIKey`
        """
        api_item = self._gis.content.add(
            {
                'title' : title,
                'type' : "API Key",
                'tags' : tags,
                'description' : description or "",
            }
        )
                
        result = api_item.register(app_type='apikey', 
                                   redirect_uris=redirect_uris, 
                                   http_referers=http_referers, 
                                   privileges=privileges)        
        return APIKey(item=api_item, gis=self._gis)
    #----------------------------------------------------------------------
    def validate(self, api_key, privileges=None):
        """
        Checks if an API has a specific privilege.
        
        ================  ===============================================================================
        **Parameter**     **Description**
        ----------------  -------------------------------------------------------------------------------
        api_key           Required `APIKey`.  The key to validate against.
        ----------------  -------------------------------------------------------------------------------
        privileges        Optional List. The list of the privileges to check for.  The list consists of 
                          a list of string values. 
        ================  ===============================================================================
        
        :returns: bool
        
        """
        if isinstance(privileges, (list, tuple)):
            privileges = ",".join(privileges)
        
        url = f"{self._base_url}oauth2/validateApiKey"
        params = {
            "f" : "json",
            "key" : api_key.properties.apiKey,
            "privilege" : privileges
        }
        res = self._gis._con.post(url, params)    
        if 'valid' in res:
            return res['valid']
        return res
    #----------------------------------------------------------------------
    @property
    def keys(self):
        """
        Returns a tuple of API Keys Registered with the Organization
        
        :returns: tuple
        
        """
        url = f"{self._base_url}portals/self/apiKeys"
        start = 0
        params = {
            "f" : "json",
            "num" : 100,
            "start" : start
        }
        from .. import Item
        res = self._gis._con.post(url, params)
        k = [APIKey(Item(gis=self._gis, itemid=k['itemId']), 
                    gis=self._gis) \
             for k in res["apiKeys"]]
        while res['nextStart'] > 0:
            start += 100
            params['start'] = start
            res = self._gis._con.post(url, params)
            k.extend(k = [APIKey(Item(gis=self._gis, itemid=k['itemId']), 
                                 gis=self._gis) \
                          for k in res["apiKeys"]])
            if res['nextStart'] <= 0:
                break
        return tuple(k)


from arcgis.gis import GIS
from arcgis._impl.common._mixins import PropertyMap
from ._base import BaseMissionServer
from ._logs import LogManager
from ._system import SystemManager
from ._machines import MachineManager
from ._security import SecurityManager
###########################################################################
class MissionServer(BaseMissionServer):
    """
    A Mission Server Instance.
    """
    _url = None
    _gis = None
    _con = None
    _properties = None
    _machinemgr = None
    _securitymgr = None 
    _security = None
    _system = None
    _logs = None
    _machine = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis=None):
        if url.lower().find("/admin") == -1:
            if url.endswith('/'):
                url = url[:-1]
            url += "/admin/"
        self._url = url
        super().__init__(url, gis)
        if gis is None:
            from arcgis import env
            gis = env.active_gis
        if gis is None:
            raise ValueError("A GIS could not be obtained.")
        self._gis = gis
        self._con = self._gis._con        
    #----------------------------------------------------------------------
    @property
    def info(self):
        """
        Returns information about the server site itself

        :returns: PropertyMap

        """
        url = self._url + "/info"
        params = {'f' : 'json'}
        res = self._gis._con.get(url, params)
        return PropertyMap(res)    
    #----------------------------------------------------------------------
    @property
    def logs(self):
        """
        Provides access to the Mission server's logging system

        :returns: LogManager

        """
        if self._logs is None:
            url = self._url + "/logs"
            self._logs = LogManager(url=url, gis=self._gis)
        return self._logs
    #----------------------------------------------------------------------
    @property
    def system(self):
        """
        returns access to the system properties of the ArcGIS Mission Server

        :return: SystemManager

        """
        if self._system is None:
            url = self._url + "/system"
            self._system = SystemManager(url=url, gis=self._gis)
        return self._system
    #----------------------------------------------------------------------
    @property
    def machine(self):
        """
        Provides access to managing the registered machines with ArcGIS
        Mission Server

        :returns: MachineManager

        """
        if self._machine is None:
            url = self._url + "/machines"
            self._machine = MachineManager(url=url, gis=self._gis)
        return self._machine
    #----------------------------------------------------------------------
    @property
    def security(self):
        """
        Provides access to managing the ArcGIS Mission Server's security
        settings.

        :return: SecurityManager

        """
        if self._security is None:
            url = self._url + "/security"
            self._security = SecurityManager(url=url, gis=self._gis)
        return self._security



from .env import HAS_TENSORFLOW
if HAS_TENSORFLOW:
    from .common_tf import NormalizationLayerRGB
    from tensorflow.keras.layers import Layer, Input
    from tensorflow.keras.models import Model
    from .object_detection import get_TFOD_post_processed_model
    from .image_classification import get_TFIC_post_processed_model
    from .pixel_classification import analyze_pred_TFPC

def get_post_processed_model_tf(arcgis_model, input_normalization=True):
    if arcgis_model.__class__.__name__ == 'SingleShotDetector':
        return get_TFOD_post_processed_model(arcgis_model, input_normalization=input_normalization)
    elif arcgis_model.__class__.__name__ == 'FeatureClassifier':
        return get_TFIC_post_processed_model(arcgis_model, input_normalization=input_normalization)
    elif arcgis_model.__class__.__name__ == 'UnetClassifier':
        postprocessing_function = analyze_pred_TFPC

    model = arcgis_model.learn.model
    input_layer = model.input
    model_output = model.output

    if input_normalization:
        input_layer = Input(tuple(input_layer.shape[1:]))
        x = NormalizationLayerRGB()(input_layer)
        model_output = model(x)
    output_layer = postprocessing_function(model_output)
    new_model = Model(input_layer, output_layer)
    return new_model


import logging
import traceback
from functools import partial


HAS_FASTAI = True

try:
    import torch.nn as nn
    from fastai.torch_core import flatten_model
except Exception as e:
    import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_FASTAI = False

logger = logging.getLogger()


def split_into_layer_groups(model, architecture, task="classification"):
    """
    Method responsible for getting the correct layer
    group splitter function and calling it to split the
    transformer model into different layer groups

    =====================   =================================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------------
    model                   Required ModelObject. The transformer model for
                            which we want to get the layer groups
    ---------------------   -------------------------------------------------
    ---------------------   -------------------------------------------------
    architecture            Required string. The transformer architecture for
                            which we wish to get the layer groups. This param
                            will be used to call the correct function to split
                            model layers
    ---------------------   -------------------------------------------------
    return: A list containing model layer groups
    """
    if task == "classification":
        splitter = get_layer_group_splitter_for_classification(architecture)
    elif task == "ner":
        splitter = get_layer_group_splitter_for_ner(architecture)
    elif task == "sequence_translation":
        splitter = get_layer_group_splitter_for_sequence_translation(architecture)
    else:
        raise Exception(f"Wrong task - {task} selected. Allowed values are 'ner', 'classification','sequence_translation'")

    # logger.info(f"Invoking - {splitter.__name__} function for splitting {architecture} model into layer groups")
    return splitter(model, architecture)


def get_layer_group_splitter_for_ner(architecture):
    """
    This function will return the appropriate function which will
    then be used to split the transformer model into layer groups

    =====================   =================================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------------
    architecture            Required string. The transformer architecture for
                            which we wish to get the layer groups. This param
                            will be used to return the correct function to
                            split model layers
    ---------------------   -------------------------------------------------
    """
    if architecture in ['bert', 'roberta', 'distilbert', 'xlm-roberta', 'electra']:
        return _bert_layer_splitter_for_ner
    elif architecture == 'albert':
        return _albert_layer_splitter
    elif architecture == "xlnet":
        return _xlnet_layer_splitter_for_ner
    elif architecture in ["xlm"]:
        _xlm_layer_splitter_for_ner = partial(_xlm_layer_splitter, task="ner")
        _xlm_layer_splitter_for_ner.__name__ = "xlm_layer_splitter_for_ner"
        return _xlm_layer_splitter_for_ner
    else:
        return naive_model_splitter

def get_layer_group_splitter_for_classification(architecture):
    """
    This function will return the appropriate function which will
    then be used to split the transformer model into layer groups

    =====================   =================================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------------
    architecture            Required string. The transformer architecture for
                            which we wish to get the layer groups. This param
                            will be used to return the correct function to
                            split model layers
    ---------------------   -------------------------------------------------
    """
    if architecture in ['bert', 'roberta', 'distilbert', 'xlm-roberta', 'electra']:
        return _bert_layer_splitter_for_classification
    elif architecture == 'albert':
        return _albert_layer_splitter
    elif architecture == "xlnet":
        return _xlnet_layer_splitter_for_classification
    elif architecture in ["xlm", "flaubert"]:
        return _xlm_layer_splitter
    else:
        return naive_model_splitter

def get_layer_group_splitter_for_sequence_translation(architecture):
    """
    This function will return the appropriate function which will
    then be used to split the transformer model into layer groups

    =====================   =================================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------------
    architecture            Required string. The transformer architecture for
                            which we wish to get the layer groups. This param
                            will be used to return the correct function to
                            split model layers
    ---------------------   -------------------------------------------------
    """
    if architecture in ["t5"]:
        return _t5_conditional_generation_splitter
    elif architecture in ["bart","mbart",'marian']:
        return _bart_conditional_generation_splitter
    else:
        return naive_model_splitter


def naive_model_splitter(model, model_name):
    if not HAS_FASTAI:
        from .._data import _raise_fastai_import_error
        _raise_fastai_import_error(import_exception=import_exception)
    linear_layer_found = False
    index, last_index = -1, -1
    layers = flatten_model(model)
    for idx, layer in reversed(list(enumerate(layers))):
        if isinstance(layer, nn.modules.linear.Linear):
            if linear_layer_found is False:
                last_index = idx
                linear_layer_found = True
                continue
            else:
                current_index = idx
                index = last_index if last_index - current_index > 4 else current_index
                break

    return [nn.Sequential(*layers[:index]), nn.Sequential(*layers[index:])]


def split_into_chunks(arr, chunk_size=3):
    # split array into 3 groups
    """
    Function to split the list into equal sized chunks
    """
    # n = round(len(arr)/ 3)
    # return [arr[:n], arr[n: 2*n], arr[2*n:]]
    return [arr[i: i+chunk_size] for i in range(0, len(arr), chunk_size)]


def _bert_layer_splitter_for_ner(model, model_name):
    """
    Split BERT, RoBERTa, DistilBERT and XLM-RoBERTa Models into layer groups
    """
    # hack to handle `xlm-roberta` string name in model_name
    model_name = model_name.split("-")[-1]
    if hasattr(model, model_name):
        model_obj = getattr(model, model_name)
        embedder = model_obj.embeddings
        pooler = None if model_name in ["distilbert", "electra"] else model_obj.pooler
        layers = model_obj.transformer.layer if model_name == "distilbert" else model_obj.encoder.layer

        chunk_size = 3 if model_name == "distilbert" else 4
        chunks = split_into_chunks(layers, chunk_size=chunk_size)

        classifier = [model.dropout, model.classifier]

        if model_name in ["distilbert", "electra"]:
            groups = [[embedder], *chunks, classifier]
        else:
            groups = [[embedder], *chunks, [pooler] + classifier]
        return groups
    else:
        raise Exception("Error in splitting the model into layer groups")


def _bert_layer_splitter_for_classification(model, model_name):
    """
    Split BERT, RoBERTa, DistilBERT and XLM-RoBERTa Models into layer groups
    """
    # hack to handle `xlm-roberta` string name in model_name
    model_name = model_name.split("-")[-1]
    if hasattr(model, model_name):
        model_obj = getattr(model, model_name)
        embedder = model_obj.embeddings
        pooler = None if model_name in ["distilbert", "electra"] else model_obj.pooler
        layers = model_obj.transformer.layer if model_name == "distilbert" else model_obj.encoder.layer

        chunk_size=4
        if model_name in ['bert']:
            classifier = [model.dropout, model.classifier]
        elif model_name in ['roberta', 'xlm-roberta', "electra"]:
            classifier = [model.classifier]
        else:
            chunk_size = 3
            classifier = [model.pre_classifier, model.classifier, model.dropout]

        chunks = split_into_chunks(layers, chunk_size=chunk_size)

        if model_name in ["distilbert", "electra"]:
            groups = [[embedder], *chunks, classifier]
        else:
            groups = [[embedder], *chunks, [pooler] + classifier]
        return groups
    else:
        raise Exception("Error in splitting the model into layer groups")


def _albert_layer_splitter(model, model_name):
    """
    Split Hugging Face ALBERT Model into layer groups
    """
    if hasattr(model, model_name):
        model_obj = getattr(model, model_name)
        embedder = model_obj.embeddings
        encoder = model_obj.encoder
        pooler = [model_obj.pooler, model_obj.pooler_activation]
        classifier = [model.dropout, model.classifier]
        groups = [[embedder], [encoder], pooler + classifier]
        return groups
    else:
        raise Exception("Error in splitting the model into layer groups")


def _xlnet_layer_splitter_for_ner(model, model_name):
    """
    Split Hugging Face XLNet Model into layer groups
    """
    if hasattr(model, "transformer"):
        model_obj = getattr(model, "transformer")
        embedder = model_obj.word_embedding
        layers = model_obj.layer
        # chunks = [layers[i: i+grouping_param] for i in range(0, len(layers), grouping_param)]
        chunks = split_into_chunks(layers, chunk_size=4)
        classifier = [model_obj.dropout, model.classifier]
        groups = [[embedder], *chunks, classifier]
        return groups
    else:
        raise Exception("Error in splitting the model into layer groups")


def _xlnet_layer_splitter_for_classification(model, model_name):
    """
    Split Hugging Face XLNet Model into layer groups
    """
    if hasattr(model, "transformer"):
        model_obj = getattr(model, "transformer")
        embedder = model_obj.word_embedding
        layers = model_obj.layer
        # chunks = [layers[i: i+grouping_param] for i in range(0, len(layers), grouping_param)]
        chunks = split_into_chunks(layers, chunk_size=4)
        seq_summary = [model.sequence_summary]
        classifier = [model.logits_proj]
        groups = [[embedder], *chunks, seq_summary + classifier]
        return groups
    else:
        raise Exception("Error in splitting the model into layer groups")


def _xlm_layer_splitter(model, model_name, task="classification"):
    """
    Split Hugging Face XLM Model into layer groups
    """
    if hasattr(model, "transformer"):
        model_obj = getattr(model, "transformer")
        embedder = [model_obj.position_embeddings, model_obj.embeddings, model_obj.layer_norm_emb]

        attentions = model_obj.attentions
        attention_chunks = split_into_chunks(attentions, chunk_size=6)
        # attention_chunks = [attentions[i: i+grouping_param] for i in range(0, len(attentions), grouping_param)]

        layer_norm1 = model_obj.layer_norm1
        layer_norm1_chunks = split_into_chunks(layer_norm1, chunk_size=6)
        # layer_norm1_chunks = [layer_norm1[i: i+grouping_param] for i in range(0, len(layer_norm1), grouping_param)]

        ffns = model_obj.ffns
        ffns_chunks = split_into_chunks(ffns, chunk_size=6)
        # ffns_chunks = [ffns[i: i+grouping_param] for i in range(0, len(ffns), grouping_param)]

        layer_norm2 = model_obj.layer_norm2
        layer_norm2_chunks = split_into_chunks(layer_norm2, chunk_size=6)
        # layer_norm2_chunks = [layer_norm2[i: i+grouping_param] for i in range(0, len(layer_norm2), grouping_param)]

        if task == "classification":
            classifier = [model.sequence_summary]
        elif task == "ner":
            classifier = [model.dropout, model.classifier]
        else:
            raise Exception(f"Wrong task - {task} selected. Allowed values are 'ner', 'classification'")
        groups = [embedder, *attention_chunks, *layer_norm1_chunks, *ffns_chunks, *layer_norm2_chunks, classifier]
        return groups
    else:
        raise Exception("Error in splitting the model into layer groups")

def _t5_conditional_generation_splitter(model, model_name):
    """
    Split Hugging Face t5 Model into layer groups
    """
    try:
        t = model
        groups = [[t.encoder.block,t.decoder.block],[t.lm_head]]
        return groups
    except:
        raise Exception("Error in splitting the model into layer groups")

def _bart_conditional_generation_splitter(model, model_name):
    """
    Split Hugging Face t5 Model into layer groups
    """
    try:
        t = model.model
        groups = [[t.encoder.embed_positions,t.encoder.layers,t.decoder.embed_positions,t.decoder.layers],[t.shared]]
        return groups
    except:
        raise Exception("Error in splitting the model into layer groups")

        



import math
from .env import ARCGIS_ENABLE_TF_BACKEND
if ARCGIS_ENABLE_TF_BACKEND:
    import tensorflow as tf
    from .common_tf import get_channel_axis

try:
    import torch
    import fastai
    import numpy as np
    import matplotlib.pyplot as plt
    from fastai.vision import imagenet_stats
    HAS_FASTAI = True
except:
    HAS_FASTAI = False

def segmentation_mask_to_one_hot(
    segmentation_mask, 
    n_classes: int
    ):
    """
    returns a one hot encoding of the supplied segmentation mask

    segmentation_mask: torch.tensor with shape (batch, 1, rows, cols) or (batch, rows, cols) or (rows, cols)
    """
    class_embeddings = torch.eye(n_classes, dtype=torch.uint8)
    shp = segmentation_mask.shape
    if len(shp) == 2:
        target_reshape_shape = (shp[0], shp[1], n_classes)
        mutation = (2, 0, 1)
    else:
        target_reshape_shape = (shp[0], shp[-2], shp[-1], n_classes)
        mutation = (0, 3, 1, 2)

    return class_embeddings[segmentation_mask.flatten()].reshape(target_reshape_shape).permute(*mutation)

def analyze_pred_pixel_classification(self, activations):
    """
    Analyzes predictions to be ready to plot
    returns numpy array.
    """
    if self._backend == 'pytorch':
        if type(activations) == list:
            activations = torch.cat(activations)
        
        if self._ignore_mapped_class != []:
            for k in self._ignore_mapped_class:
                activations[:, k] = -1
        return activations.max(dim=1)[1].cpu().numpy()
    elif self._backend == 'tensorflow':
        if type(activations) == list:
            activations = tf.concat(activations, 0)
        return analyze_pred_TFPC(activations).numpy()

def analyze_pred_TFPC(activations):
    return tf.argmax(activations, axis=get_channel_axis())



import os
import datetime
import concurrent.futures
from concurrent.futures import Future
import logging
import json
_log = logging.getLogger(__name__)

class StatusJob(object):
    """
    Represents a Single Geoprocessing Job.  The `GPJob` class allows for the asynchronous operation
    of any geoprocessing task.  To request a GPJob task, the code must be called with `future=True`
    or else the operation will occur synchronously.  This class is not intended for users to call
    directly.


    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    future            Required current.futures.Future.  The async object.
    ----------------  ---------------------------------------------------------------
    op                Required String. The name of the operation performed.
    ----------------  ---------------------------------------------------------------
    jobid             Required String. The unique ID of the job.
    ----------------  ---------------------------------------------------------------
    gis               Required GIS. The GIS connection object
    ----------------  ---------------------------------------------------------------
    notify            Optional Boolean.  When set to True, a message will inform the
                      user that the geoprocessing task has completed. The default is
                      False.
    ----------------  ---------------------------------------------------------------
    extra_marker      Optional String. An extra piece of text to place infront of the 
                      Job string for the __repr__ object.
    ================  ===============================================================

    """
    _future = None
    _jobid = None
    _url = None
    _gis = None
    _task_name = None
    _is_fa = False
    _is_ra = False
    _is_ortho = False
    _start_time = None
    _end_time = None
    _item_properties = None
    #----------------------------------------------------------------------
    def __init__(self, future, op, jobid, gis, notify=False, extra_marker="Group"):
        """
        initializer
        """
        assert isinstance(future, Future)
        self._thing = extra_marker
        self._future = future
        self._start_time = datetime.datetime.now()
        if notify:
            self._future.add_done_callback(self._notify)
        self._future.add_done_callback(self._set_end_time)
        self._op = op
        self._jobid = jobid
        self._gis = gis
    #----------------------------------------------------------------------
    @property
    def ellapse_time(self) -> datetime.datetime:
        """
        Returns the Ellapse Time for the Job
        """
        if self._end_time:
            return self._end_time - self._start_time
        else:
            return datetime.datetime.now() - self._start_time
    #----------------------------------------------------------------------
    @property
    def definition(self) -> dict:
        """
        Returns information about the job

        :return: Dict
        """
        url = f"{self._gis._portal.resturl}portals/self/jobs/%s" % self._jobid
        params = {'f' : 'json'}
        res = self._gis._con.post(url, params)
        if "definition" in res:
            return res["definition"]
        return res

    #----------------------------------------------------------------------
    def _set_end_time(self, future):
        """sets the finish time"""
        self._end_time = datetime.datetime.now()
    #----------------------------------------------------------------------
    def _notify(self, future):
        """prints finished method"""
        jobid = self._jobid
        try:
            res = future.result()
            infomsg = '{jobid} finished successfully.'.format(jobid=jobid)
            _log.info(infomsg)
            print(infomsg)
        except Exception as e:
            msg = str(e)
            msg = '{jobid} failed: {msg}'.format(jobid=jobid, msg=msg)
            _log.info(msg)
            print(msg)
    #----------------------------------------------------------------------
    def __str__(self) -> str:
        return self.__repr__()
    #----------------------------------------------------------------------
    def __repr__(self) -> str:
        if len(self._thing) > 0:
            return "<%s %s Job: %s>" % (self.task, self._thing, self._jobid)
        else:
            return "<%s Job: %s>" % (self.task, self._jobid)
    #----------------------------------------------------------------------
    @property
    def task(self) -> str:
        """Returns the task name.
        :returns: string
        """
        return self._op
    #----------------------------------------------------------------------
    @property
    def status(self) -> str:
        """
        returns the GP status

        :returns: String
        """

        url = f"{self._gis._portal.resturl}portals/self/jobs/%s" % self._jobid
        params = {'f' : 'json'}
        res = self._gis._con.post(url, params)
        if "status" in res:
            return res["status"]
        return res
    #----------------------------------------------------------------------
    @property
    def messages(self) -> list:
        """
        returns the jobs message

        :returns: String
        """

        url = f"{self._gis._portal.resturl}portals/self/jobs/%s" % self._jobid
        params = {'f' : 'json'}
        res = self._gis._con.post(url, params)
        if "messages" in res:
            return res["messages"]
        return res
    #----------------------------------------------------------------------
    def cancel(self) -> bool:
        """
        Cancels the `Future` process to end the job locally.
        Import/Export jobs cannot be terminiated on server.

        :returns: boolean
        """
        if self.done():
            return False
        if self.cancelled():
            return False
        self._future.cancel()
        return True
    #----------------------------------------------------------------------
    def cancelled(self) -> bool:
        """
        Return True if the call was successfully cancelled.

        :returns: boolean
        """
        return self._future.cancelled()
    #----------------------------------------------------------------------
    def running(self) -> bool:
        """
        Return True if the call is currently being executed and cannot be cancelled.

        :returns: boolean
        """
        return self._future.running()
    #----------------------------------------------------------------------
    def done(self) -> bool:
        """
        Return True if the call was successfully cancelled or finished running.

        :returns: boolean
        """
        return self._future.done()
    #----------------------------------------------------------------------
    def result(self):
        """
        Return the value returned by the call. If the call hasn't yet completed
        then this method will wait.

        :returns: object
        """
        from arcgis.gis import Item
        if self.cancelled():
            return None
        res = self._future.result()
        if 'result' in res:
            if 'itemId' in res['result']:
                from arcgis.gis import Item
                return Item(gis=self._gis, itemid=res['result']['itemId'])
            elif 'itemsImported' in res['result']:
                return_result = {}
                return_result['itemsImported'] = [Item(itemid=i['itemId'], gis=self._gis) for i in res['result']['itemsImported'] if 'itemId' in i]
                return_result['itemsSkipped'] = [Item(i['itemId'], gis=self._gis) for i in res['result']['itemsSkipped'] if 'itemId' in i]
                return_result['itemsFailedImport'] = [Item(i['itemId'], gis=self._gis) for i in res['result']['itemsFailedImport'] if 'itemId' in i]
                return return_result
            elif 'services' in res['result']:
                return [Item(self._gis, t['serviceItemId']) for t in res['result']['services'] if 'serviceItemId' in t]            
            else:
                return res
            
        return res



import os
import sys
import json
import uuid
import tempfile
from arcgis.gis import GIS, Item
from arcgis import env as _env
import pandas as pd
from ._base import BaseOpenData
_PD_LESS_THAN1 = [int(v) for v in pd.__version__.split(".")] < [1,0,0]
###########################################################################
class CSVLayer(BaseOpenData):
    r"""
    Represents a CSV File Hosted on a Server.


    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    url_or_item         Required String or Item. The web address or `Item` to the CSV resource.
    ---------------     --------------------------------------------------------------------
    gis                 Optional GIS. The GIS used to reference the service. The arcgis.env.active_gis is used if not specified.
    ---------------     --------------------------------------------------------------------
    copyright           Optional String. Describes limitations and usage of the data.
    ---------------     --------------------------------------------------------------------
    delimiter           Optional String. The separator value. This can be the following:
                        , (comma), ' ' (space), | (pipe), \\r (tab), or ; (semicolon).
    ---------------     --------------------------------------------------------------------
    fields              Optional List. An array of dictionarys containing the field information.
    ---------------     --------------------------------------------------------------------
    opacity             Optional Float.  This value can range between 1 and 0, where 0 is 100 percent transparent and 1 is completely opaque.
    ---------------     --------------------------------------------------------------------
    scale               Optional Tuple. The min/max scale of the layer where the positions are: (min, max) as float values.
    ---------------     --------------------------------------------------------------------
    sql_expression      Optional String. Optional query string to apply to the layer when displayed on the widget or web map.
    ---------------     --------------------------------------------------------------------
    title               Optional String. The title of the layer used to identify it in places such as the Legend and Layer List widgets.
    ===============     ====================================================================

    """
    _url = None
    _gis = None
    _data = None
    _nrows = 15 #default number of rows to peak at.
    _id = None
    _renderer = None
    _latitude = None
    _longitude = None
    _type = "CSV"
    #----------------------------------------------------------------------
    def __init__(self, url_or_item, gis=None, **kwargs):
        """initializer"""
        super(CSVLayer, self)
        if isinstance(url_or_item, str):
            self._url = url_or_item
            self._item = None
        elif isinstance(url_or_item, Item):
            self._item = url_or_item
            self._url = None
        self._gis = gis or _env.active_gis or self._item._gis or GIS()
        self._copyright = kwargs.pop('copyright', None)
        self._delimiter = kwargs.pop('delimiter', ',')
        self._fields = kwargs.pop('fields', None)
        self._sql = kwargs.pop('sql_expression', None)
        self._id = kwargs.pop('id', uuid.uuid4().hex)
        self._title = kwargs.pop('title', 'CSV Layer')
        self._min_scale, self._max_scale = kwargs.pop('scale', (0,0))
        self._opacity = kwargs.pop('opacity', 1)
    #----------------------------------------------------------------------
    def __str__(self):
        if self._item:
            return f"<CSV @ {self._item.itemid}>"
        return f"<CSV @ {self._url}>"
    #----------------------------------------------------------------------
    @property
    def latitude(self):
        """
        The latitude field name. If not specified, the class will look for
        following field names in the CSV source: "lat", "latitude",
        "y", "ycenter", "latitude83", "latdecdeg", "POINT-Y".
        """
        auto_lat = ["lat", "latitude", "y",
                    "ycenter", "latitude83",
                    "latdecdeg", "point-y"]
        if self._latitude is None:
            for f in self.fields:
                if f['name'].lower() in auto_lat:
                    self._latitude = f['name']
                    break
        return self._latitude
    #----------------------------------------------------------------------
    @latitude.setter
    def latitude(self, value):
        if value != self._latitude and \
           value in [f['name'] for f in self.fields]:
            self._latitude = value
    #----------------------------------------------------------------------
    @property
    def longitude(self):
        """
        The longitude field name. If not specified, the `CSVLayer` will
        look for following field names in the CSV source: "lon", "lng",
        "long", "longitude", "x", "xcenter", "longitude83", "longdecdeg",
        "POINT-X".
        """
        auto_lat = ["lon", "lng",
                    "long", "longitude",
                    "x", "xcenter",
                    "longitude83", "longdecdeg",
                    "point-x"]
        if self._longitude is None:
            for f in self.fields:
                if f['name'].lower() in auto_lat:
                    self._longitude = f['name']
                    break
        return self._longitude
    #----------------------------------------------------------------------
    @longitude.setter
    def longitude(self, value):
        if value != self._longitude and \
           value in [f['name'] for f in self.fields]:
            self._longitude = value
    #----------------------------------------------------------------------
    @property
    def renderer(self):
        """
        Get/Set the Renderer of the CSV Layer

        :returns: InsensitiveDict

        """
        from arcgis._impl.common._isd import InsensitiveDict
        if self._renderer is None:
            from arcgis.mapping import generate_renderer
            sr = generate_renderer(geometry_type="point")
            self._renderer = InsensitiveDict(dict(sr))
        return self._renderer
    #----------------------------------------------------------------------
    @renderer.setter
    def renderer(self, value):
        """
        Get/Set the Renderer of the CSV Layer

        :returns: InsensitiveDict

        """
        from arcgis._impl.common._isd import InsensitiveDict
        if isinstance(value, dict):
            self._renderer = InsensitiveDict(dict(value))
        elif value is None:
            self._renderer = None
        elif not isinstance(value, InsensitiveDict):
            raise ValueError("Invalid renderer type.")
        self._refresh = value

    #----------------------------------------------------------------------
    @property
    def delimiter(self):
        r"""
        Gets/Sets the delimiter for the CSV Layer.  The default is `,`

        ===========   ==========================================
        **Values**    **Description**
        -----------   ------------------------------------------
        ,             Comma
        -----------   ------------------------------------------
        " "           space
        -----------   ------------------------------------------
        ;             semicolon
        -----------   ------------------------------------------
        |             pipe
        -----------   ------------------------------------------
        `\r`          tab
        ===========   ==========================================

        :returns: string

        """
        if self._delimiter is None:
            self._delimiter = ","
        return self._delimiter
    #----------------------------------------------------------------------
    @delimiter.setter
    def delimiter(self, value):
        r"""
        Gets/Sets the delimiter for the CSV Layer.  The default is `,`

        ===========   ==========================================
        **Values**    **Description**
        -----------   ------------------------------------------
        ,             Comma
        -----------   ------------------------------------------
        " "           space
        -----------   ------------------------------------------
        ;             semicolon
        -----------   ------------------------------------------
        |             pipe
        -----------   ------------------------------------------
        `\r`          tab
        ===========   ==========================================

        :returns: string

        """
        if value in [',', ' ', ';', '|', '\r'] and \
           self._delimiter != value:
            self._delimiter = value
    #----------------------------------------------------------------------
    @property
    def fields(self):
        """
        Returns the fields values for the CSV source.

        :returns: list of strings
        """
        import numpy as np
        import datetime
        if self._data is None and self._fields is None:
            fields = []
            self._data = self._df(True)
            for col in self._data.columns:
                try:
                    idx = self._data[col].first_valid_index()
                    col_val = self._data[col].loc[idx]
                except:
                    col_val = ""
                if isinstance(col_val, (str, np.str)):
                    fields.append({
                        "name" : col,
                        "type" : "string",
                        "alias" : col
                    })
                elif isinstance(col_val, (datetime.datetime,
                                          pd.Timestamp,
                                          np.datetime64,
                                          )):
                    fields.append({
                        "name" : col,
                        "type" : "date",
                        "alias" : col
                    })
                elif isinstance(col_val, (np.int32, np.int16, np.int8)):
                    fields.append({
                        "name" : col,
                        "type" : "long",
                        "alias" : col
                    })
                elif isinstance(col_val, (int, np.int, np.int64)):
                    fields.append({
                        "name" : col,
                        "type" : "integer",
                        "alias" : col
                    })
                elif isinstance(col_val, (float, np.float64)):
                    fields.append({
                        "name" : col,
                        "type" : "double",
                        "alias" : col
                    })
                elif isinstance(col_val, (np.float32)):
                    fields.append({
                        "name" : col,
                        "type" : "single",
                        "alias" : col
                    })
            self._fields = fields
        return self._fields
    #----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        """Represents the MapView's JSON format"""
        add_layer =  {
            "type" : self._type,
            "delimiter" : self.delimiter,
            "copyright" : self.copyright or "",
            "definitionExpression" : self.sql_expression or "",
            'layerDefinition' : {
                'fields' : self.fields,
                'objectIDField': "__OBJECTID",
                'drawingInfo': {
                    "renderer" : self.renderer._json()}},
            'id' : self._id,
            'title' : self.title,
            'opacity' : self.opacity,
            'maxScale' : self.scale[1],
            'minScale' : self.scale[0],
            'locationInfo' : {
                'locationType' : 'coordinates',
                'longitudeFieldName' : self.longitude,
                'latitudeFieldName' : self.latitude,
            }
        }
        if self._item:
            add_layer["portalItem"] = { "id" : self._item.itemid }
        else:
            add_layer["url"] = self._url
        return add_layer

    @property
    def _operational_layer_json(self):
        """Represents the WebMap's JSON format"""
        return self._lyr_json

    #----------------------------------------------------------------------
    def _df(self, glance=False):
        """returns the data as a pd.DataFrame"""
        if glance:
            nrows = self._nrows
        else:
            nrows = None
        if self._url:
            return pd.read_csv(self._url,
                                   sep=self.delimiter,
                                   nrows=nrows,
                                   infer_datetime_format=True,
                                   parse_dates=True)
        elif self._item:
            if self._item._gis._con.token:
                url = f"{self._item._gis._portal.resturl}content/items/{self._item.itemid}/data?token={self._item._gis._con.token}"
            else:
                url = f"{self._item._gis._portal.resturl}content/items/{self._item.itemid}/data"
            return pd.read_csv(url,
                               sep=self.delimiter,
                               nrows=nrows,
                               infer_datetime_format=True,
                               parse_dates=True)
        else:
            return None
    #----------------------------------------------------------------------
    @property
    def df(self):
        """
        returns the CSV file as a DataFrame

        :returns: Pandas' DataFrame
        """
        return self._df(False)





import random
import torch
from torch import nn
import math
import numpy as np
import torch.nn.functional as F
from fastai.basic_train import LearnerCallback
from torch_geometric.nn import fps, knn

# For AverageMetric callback.
from fastai.callback import Callback
from fastai.core import first_el , is_listy
from fastai.torch_core import add_metrics, num_distrib
import torch.distributed as dist

def farthest_point_sample(pts, npoint):
    """
    Input:
        xyz: pointcloud data, [B, N, C]
        npoint: number of samples
    Return:
        centroids: sampled pointcloud index, [B, npoint]
    """
    device = pts.device
    B, N, C = pts.shape
    
    batch = (torch.arange(0, B*N) // N).to(pts.device).contiguous()
    pts = pts.view(-1, C).float().contiguous()
    indices = fps(pts, batch, ratio=(npoint/N))
    if (indices.shape[0] / B) > npoint:
        each_block_point = int(indices.shape[0] / B)
        pts = pts[indices].view(B, each_block_point, C)
        drop_point_num = each_block_point - npoint
        pts = pts[:, :-drop_point_num, :]
    else:
        pts = pts[indices].view(B, npoint, C) 
    return pts.contiguous()

def find_k_neighbor(rep_pts, pts, K, D):
    """
    Find K nearest neighbor points
    :param pts: represent points(B, P, C)
    :param rep_pts: original points (B, N, C)
    :param K:
    :param D: Dilation rate
    :return group_pts: K neighbor points(B, P, K, C)
    """
    device = pts.device
    B, N, C = pts.shape
    _, N_rep, _ = rep_pts.shape
    batch_pts = (torch.arange(0, B*N) // N).to(pts.device).contiguous()
    batch_rep_pts = (torch.arange(0, B*N_rep) // N_rep).to(pts.device).contiguous()
    pts = pts.view(-1, C).contiguous()
    rep_pts = rep_pts.view(-1, C).contiguous()
    knn_indices = knn(pts, rep_pts, K * D, batch_pts, batch_rep_pts)
    knn_indices = knn_indices[1] ## grab the indices at 0th index it is [0,0,0,0,0,1,1,1,1,1,2,2,2,2,...]  
    group_pts = pts[knn_indices].view(B, N_rep, K * D, C).contiguous()
    rand_col = torch.randint(K * D, (K,))
    group_pts = group_pts[:, :, rand_col, :]

    return group_pts.contiguous(), knn_indices.contiguous(), rand_col.contiguous()

class DepthwiseConv2D(nn.Module):
    def __init__(self, in_channels, depth_multiplier, kernel_size, final_reshape=False, point_wise=None):
        """
        in_channels: input channels of point cloud,
        depth_multiplier: in nn.Conv2d out_channels is set to depth_multiplier * in_channels and groups=in_channels
        final_reshape: returns after reshaping and permuting the tensor
        point_wise: accepts a integer which is the out channel of the XConv
        """
        super(DepthwiseConv2D, self).__init__()
        self.depth_multiplier = depth_multiplier
        self.final_reshape = final_reshape
        self.point_wise = bool(point_wise)
        if point_wise:
            self.module = nn.Sequential(nn.Conv2d(in_channels=in_channels,
                                out_channels=depth_multiplier * in_channels,
                                kernel_size=kernel_size,
                                groups=in_channels,
                                bias=False),
                                nn.Conv2d(in_channels=depth_multiplier * in_channels,
                                          out_channels=point_wise,
                                          kernel_size=1,
                                          bias=False
                                         ),
                                nn.ELU(inplace=True),
                                nn.BatchNorm2d(point_wise)
                                )
        else:
            self.module = nn.Sequential(nn.Conv2d(in_channels=in_channels,
                                out_channels=depth_multiplier * in_channels,
                                kernel_size=kernel_size,
                                groups=in_channels,
                                bias=False),
                                nn.ELU(inplace=True),
                                nn.BatchNorm2d(depth_multiplier * in_channels)
                                )
        
        
    def forward(self, inp):
        if self.point_wise:
            out = self.module(inp.contiguous())
            if self.final_reshape:
                return out.squeeze(3).permute(0, 2 ,1).contiguous()
            else:
                return out
        else:
            B, C, P, K = inp.shape
            C = int(C ** 0.5)
            inp = inp.view(B, P, C, C).permute(0, 3, 1, 2).contiguous()
            out = self.module(inp.contiguous())
            return out.view(B, P, self.depth_multiplier, C).permute(0, 3, 1, 2).contiguous() if self.final_reshape else out

class XConvDepthwise(nn.Module):
    def __init__(self, in_channel, lift_channel, out_channel, P, K, D=1, sampling='fps', with_global=False):
        """
        :param in_channel: Input channel of the points' features
        :param lift_channel: Lifted channel C_delta
        :param out_channel:
        :param P: P represent points
        :param K: K neighbors to operate
        :param D: Dilation rate
        """
        super(XConvDepthwise, self).__init__()
        self.P = P
        self.K = K
        self.D = D
        self.sampling = sampling
        self.with_global = with_global
        
        # Input should be (B, 3, P, K)
        self.MLP_delta = nn.Sequential(
            nn.Conv2d(3, lift_channel, kernel_size=1),
            nn.ELU(inplace=True),
            nn.BatchNorm2d(lift_channel),
            nn.Conv2d(lift_channel, lift_channel, kernel_size=1),
            nn.ELU(inplace=True),
            nn.BatchNorm2d(lift_channel)
        )
        
        # Input should be (B, 3, P, K)  --> (B, K, P, K) 
        self.MLP_X = nn.Sequential(
            nn.Conv2d(3, K * K, kernel_size=(1, K)),
            nn.ELU(inplace=True),
            nn.BatchNorm2d(K * K),
            DepthwiseConv2D(in_channels=K, depth_multiplier=K, kernel_size=(1, K)),
            DepthwiseConv2D(in_channels=K, depth_multiplier=K, kernel_size=(1, K), final_reshape=True)
        )
        
        depth_multiplier = math.ceil(out_channel / (in_channel + lift_channel))
        self.seperable_conv = DepthwiseConv2D(in_channel + lift_channel,
                                              depth_multiplier=depth_multiplier,
                                              kernel_size=(1, K),
                                              point_wise=out_channel,
                                              final_reshape=True
                                             )

        if with_global:
            lift_channel = out_channel // 4
            self.MLP_g = nn.Sequential(
                nn.Conv2d(3, lift_channel, kernel_size=1),
                nn.ELU(inplace=True),
                nn.BatchNorm2d(lift_channel),
                nn.Conv2d(lift_channel, lift_channel, kernel_size=1),
                nn.ELU(inplace=True),
                nn.BatchNorm2d(lift_channel)
            )

    def forward(self, pts, fts, represent_pts=None):
        """
        :param x: (rep_pt, pts, fts) where
          - pts: Regional point cloud (B, N, 3)
          - fts: Regional features (B, N, C)
        :return: Features aggregated into point rep_pt.
        """
        B, N, _ = pts.size()
        
        if represent_pts is None:
            if (self.P == -1) or (self.P == N):
                self.P = N
                represent_pts = pts
            else:
                represent_pts = farthest_point_sample(pts, self.P)        # (B, P, 3)

        group_pts, k_ind, rand_col = find_k_neighbor(represent_pts, pts, self.K, self.D)  # (B, P, K, 3), (B, P, K)
        center_pts = torch.unsqueeze(represent_pts, dim=2)      # (B, P, 1, 3)
        group_pts = group_pts - center_pts       # (B, P, K, 3)
        
        # fts_lifted
        group_pts = group_pts.permute(0,3,1,2).contiguous()
        fts_lifted = self.MLP_delta(group_pts.contiguous())      # (B, C_delta, P, K)
        
        if fts is not None:
            _, _, nf = fts.shape
            group_fts = fts.contiguous().view(-1, nf)
            group_fts = group_fts[k_ind].view(B, self.P, self.K * self.D, nf)
            group_fts = group_fts[:, :, rand_col, :]
            group_fts = group_fts.permute(0, 3, 1, 2).contiguous()
            feat = torch.cat((fts_lifted, group_fts), 1).contiguous()  # (B, C_delta + C_in, P, K)
        else:
            feat = fts_lifted.contiguous()
            
        # XConv operation
        X = self.MLP_X(group_pts).permute(0,2,3,1)  # (B, P, K, K)
        X = X.contiguous().view(B*self.P, self.K, self.K)
        
        feat = feat.permute(0,2,3,1).contiguous().view(B*self.P, self.K, -1)
        feat = torch.bmm(X, feat).view(B, self.P, self.K, -1).permute(0,3,1,2)  # (B, C_delta + C_in, P, K)
        
        feat = self.seperable_conv(feat.contiguous())           # (B, self.P, C_out)
        
        if self.with_global:
            feat_global = self.MLP_g(represent_pts.unsqueeze(dim=2).permute(0,3,1,2).contiguous()).squeeze(3).permute(0,2,1).contiguous()
            return represent_pts.contiguous(), torch.cat([feat, feat_global], dim=-1).contiguous() # (B, P, 3), (B, P, C_out + C_out / 4)
        else:
            return represent_pts.contiguous(), feat.contiguous()      # (B, P, 3), (B, P, C_out)


def get_indices(batch_size, sample_num, point_num, pool_setting=None):
    if not isinstance(point_num, np.ndarray):
        point_nums = np.full((batch_size), point_num.cpu())
    else:
        point_nums = point_num

    indices = []
    for i in range(batch_size):
        pt_num = point_nums[i]
        if pool_setting is None:
            pool_size = pt_num
        else:
            if isinstance(pool_setting, int):
                pool_size = min(pool_setting, pt_num)
            elif isinstance(pool_setting, tuple):
                pool_size = min(random.randrange(pool_setting[0], pool_setting[1]+1), pt_num)
        if pool_size > sample_num:
            choices = np.random.choice(pool_size, sample_num, replace=False)
        else:
            choices = np.concatenate((np.random.choice(pool_size, pool_size, replace=False),
                                      np.random.choice(pool_size, sample_num - pool_size, replace=True)))
        if pool_size < pt_num:
            choices_pool = np.random.choice(pt_num, pool_size, replace=False)
            choices = choices_pool[choices]
        choices = np.expand_dims(choices, axis=1)
        choices_2d = np.concatenate((np.full_like(choices, i), choices), axis=1)
        indices.append(choices_2d)
    return np.stack(indices)

class SamplePointsCallback(LearnerCallback):
    def __init__(self, learn, sample_point_num):
        super().__init__(learn)
        self.sample_point_num = sample_point_num
        
    def on_batch_begin(self, last_input, last_target, **kwargs):
        """
        Random Sample points from input and output.
        """
        last_input, point_nums = last_input   ## (batch, total_points, num_features), (batch,)
        batch, _, num_features = last_input.shape
        
        ## get indices to sample (tf implementation function)
        indices = torch.tensor(get_indices(batch, self.sample_point_num, point_nums.long())).to(last_input.device)
        
        ## Get indices in the correct shape to be used for indexing
        indices = indices.view(-1, 2).long()

        ##  sample points from all the input and output points
        last_input = last_input[indices[:, 0], indices[:, 1]].view(batch, self.sample_point_num, num_features).contiguous()  ## batch, self.sample_point_num, num_features                
        last_target = last_target[indices[:, 0], indices[:, 1]].view(batch, self.sample_point_num).contiguous() ## batch, self.sample_point_num
        
        del indices

        if self.learn.data.pc_type == 'PointCloud_TF':
            if self.learn.data.transform_fn is not None and self.learn.model.training:
                if random.random() > 0.5:
                    last_input[:, :, :3] = self.learn.data.transform_fn(last_input)  
        
        return {'last_input':last_input.contiguous(), 'last_target':last_target.contiguous()}


class PointCNNSeg(nn.Module):
    def __init__(self, sampled_num_points,
                 num_classes,
                 num_extra_features,
                 encoder_params = None,
                 dropout=None,
                 ):
        """
        sampled_num_points: The number that goes into the model.
        num_classes: number of classes
        num_extra_features: number of features other than XYZ
        encoder_params: Length of out_channels, P, K, D should be same. The length denotes the number of layers in encoder.
        {
                'out_channels': Number of channels in each layer multiplied by m,
                 'P': Number of points in each layer,
                 'K': Number of K-nearest neighbour in each layer,
                 'D': Dilation in each layer,
                 'm': Multiplier which is multiplied by each out_channel.
        }
        """
        super().__init__()
        
        if encoder_params is None:
            encoder_params = {'out_channels':[16, 32, 64, 96],
                 'P':[-1, 768, 384, 128],
                 'K':[12, 16, 16, 16],
                 'D':[1, 1, 2, 2],
                 'm':8
                 }

        if dropout is None:
            dropout = 0.5

        self.num_extra_features = num_extra_features
        self.num_classes = num_classes
        self.encoder_params = encoder_params
        
        ## m is the multiplier value as on case of config file of yangli/pointcnn
        ## This value is multiplied by the out_channels value
        out_channels, P, K, D, m = list(encoder_params.values())
        
        ############## Encoder ################
        self.encoder_layers = nn.ModuleList([])
        for i in range(len(P)):
            
            self.encoder_layers.append(
                ## in_channels is equal to num_extra_features for the first layer.
                XConvDepthwise(in_channel=num_extra_features if i == 0 else out_channels[i-1] * m,
                            lift_channel=out_channels[i] * m // 4,
                            out_channel=out_channels[i] * m,
                            P=P[i],
                            K=K[i],
                            D=D[i],
                            with_global=True if i == (len(P) - 1) else False
                           )
            )
            
            
        ############## Decoder ################
        self.decoder_layers = nn.ModuleList([])
        self.densecat_layers = nn.ModuleList([])
        for j in reversed(range(len(P) - 1)):
            
            if P[j] == -1:
                P[j] = sampled_num_points
            
            self.decoder_layers.append(                                                   ## append decoder layers
                ## Since in the 
                XConvDepthwise(in_channel=out_channels[j + 1] * m + (out_channels[j + 1] * m) // 4 if (j+1) == (len(P) - 1) else out_channels[j+1] * m,
                               lift_channel=out_channels[j] * m // 4,
                               out_channel=out_channels[j] * m,
                               P=P[j],
                               K=K[j],
                               D=D[j]
                )
            )
            
            self.densecat_layers.append(                                                   ## append concat layers
                nn.Sequential(nn.Conv1d(out_channels[j] * m * 2, out_channels[j] * m, 1),
                                 nn.ELU(inplace=True),
                                 nn.BatchNorm1d(out_channels[j] * m)
                                ))
                
        self.fc_layers = nn.Sequential(
             nn.Conv1d(out_channels[j] * m, 256, 1),
             nn.ELU(inplace=True),
             nn.BatchNorm1d(256),
             nn.Dropout(dropout / 2),
             nn.Conv1d(256, 256, 1),
             nn.ELU(inplace=True),
             nn.BatchNorm1d(256),
             nn.Dropout(dropout),
             nn.Conv1d(256, num_classes, 1)
        )
        
    def forward(self, pts):
        """
        pts: points tensor of the following shape(Batch, sampled_num_points, 3 + num_extra_features)
        """

        if pts.shape[2] > 3:
            pts , fts = pts[:, :, :-self.num_extra_features], pts[:, :, -self.num_extra_features:]
        else:
            fts = None
            
        rep_pts_fts = []        
        
        for i, module in enumerate(self.encoder_layers):
            rep_pts_fts.append(module(*((pts, fts) if i==0 else rep_pts_fts[-1])))
            
        for i, module in enumerate(self.decoder_layers):
            if i==0:
                # If i==0, we do not perform densecat operation.
                _, fts_final = module(*rep_pts_fts[-(i+1)], represent_pts=rep_pts_fts[-(i+2)][0])
            else:
                dc_module = self.densecat_layers[i-1]
                ## Here we are performing the densecat operation with the earlier features of the pointcnn model.
                fts_final = dc_module(torch.cat((fts_final, rep_pts_fts[-(i+1)][1]), dim=-1).permute(0,2,1).contiguous()).permute(0,2,1).contiguous()
                ## The concatenated features from are then passed throught the XConv module.
                _, fts_final = module(rep_pts_fts[-(i+1)][0], fts_final, represent_pts=rep_pts_fts[-(i+2)][0])
        
        ## The final concatenation happens here.
        fts_final = self.densecat_layers[-1](torch.cat((fts_final, rep_pts_fts[0][1]), dim=-1).permute(0,2,1))
        
        ## Model is then passed through the fully connected layers to get the logits.
        return self.fc_layers(fts_final).permute(0,2,1)

class CrossEntropyPC(nn.Module):
    def __init__(self, num_classes):
        super().__init__()
        self.num_classes = num_classes
    
    def forward(self, inp, target):
        inp = inp.contiguous()
        target = target.contiguous()
        inp = inp.view(-1, self.num_classes).contiguous()
        target = target.view(-1).contiguous()
        return F.cross_entropy(inp, target).contiguous()

def accuracy(pred, target):
    pred = pred.contiguous()
    target = target.contiguous()    
    pred = pred.argmax(dim=-1)
    return (pred == target).float().mean()

def accuracy_non_zero(pred, target):
    pred = pred.contiguous()
    target = target.contiguous()
    pred = pred.argmax(dim=-1)
    mask = target != 0
    accuracy_value = (pred[mask] == target[mask]).float().mean()
    return accuracy_value

## Redefines AverageMetric Callback so that it handle nan values and does not compute average on those.
class AverageMetric(Callback):
    "Wrap a `func` in a callback for metrics computation."
    def __init__(self, func):
        # If func has a __name__ use this one else it should be a partial
        name = func.__name__ if hasattr(func, '__name__') else func.func.__name__
        self.func, self.name = func, name
        self.world = num_distrib()

    def on_epoch_begin(self, **kwargs):
        "Set the inner value to 0."
        self.val, self.count = 0.,0

    def on_batch_end(self, last_output, last_target, **kwargs):
        "Update metric computation with `last_output` and `last_target`."
        if not is_listy(last_target): last_target=[last_target]
        val = self.func(last_output, *last_target)
        ## If nan do not increase counter and return
        if torch.isnan(val).tolist():
            return
        self.count += first_el(last_target).size(0)
        if self.world:
            val = val.clone()
            dist.all_reduce(val, op=dist.ReduceOp.SUM)
            val /= self.world
        self.val += first_el(last_target).size(0) * val.detach().cpu()

    def on_epoch_end(self, last_metrics, **kwargs):
        "Set the final result in `last_metrics`."
        if self.count == 0:
            return add_metrics(last_metrics, [None])
        return add_metrics(last_metrics, self.val/self.count)

## Iteration Stop Callback, i.e stops epoch after certain number of iterations.
class IterationStop(LearnerCallback):
    def __init__(self, learn, stop_iteration):
        super().__init__(learn)
        self.stop_iteration = stop_iteration

    def on_batch_end(self, **kwargs):
        if (kwargs['iteration'] + 1) % self.stop_iteration==0:
            return {'stop_epoch': True}


from fastai.vision import ImageSegment, Image
from fastai.vision.image import open_image, show_image, pil2tensor
from fastai.vision.data import SegmentationProcessor, ImageList
from fastai.layers import CrossEntropyFlat
from fastai.basic_train import LearnerCallback
import torch
import warnings
import PIL
import numpy as np
from skimage import io
import matplotlib.pyplot as plt
from torch import LongTensor
import os
from .._utils.common import ArcGISMSImage

from fastprogress.fastprogress import progress_bar


class ArcGISImageSegment(Image):
    "Support applying transforms to segmentation masks data in `px`."
    def __init__(self, x, cmap=None, norm=None):
        super(ArcGISImageSegment, self).__init__(x)
        self.cmap = cmap
        self.mplnorm = norm
        self.type = np.unique(x)

    def lighting(self, func, *args, **kwargs):
        return self

    def refresh(self):
        self.sample_kwargs['mode'] = 'nearest'
        return super().refresh()

    @property
    def data(self):
        "Return this image pixels as a `LongTensor`."
        return self.px.long()

    def show(self, ax = None, figsize = (3,3), title = None, hide_axis = True, cmap='tab20', alpha = 0.5, **kwargs):

        if ax is None: fig,ax = plt.subplots(figsize=figsize)
        masks = self.data[0].numpy()
        for i in range(1, self.data.shape[0]):
            max_unique = np.max(np.unique(masks))
            mask = np.where(self.data[i]>0, self.data[i] + max_unique, self.data[i])
            masks += mask
        ax.imshow(masks, cmap=cmap, alpha=alpha, **kwargs)
        if hide_axis: ax.axis('off')
        if title: ax.set_title(title)


def is_no_color(color_mapping):
    if isinstance(color_mapping, dict):
        color_mapping = list(color_mapping.values())
    return (np.array(color_mapping) == [-1., -1., -1.]).any()

class ArcGISSegmentationLabelList(ImageList):
    "`ItemList` for segmentation masks."
    _processor = SegmentationProcessor
    def __init__(self, items, chip_size, classes=None, class_mapping=None, color_mapping=None, index_dir=None, **kwargs):
        super().__init__(items, **kwargs)
        self.class_mapping = class_mapping
        self.color_mapping = color_mapping
        self.copy_new.append('classes')
        self.classes, self.loss_func = classes, CrossEntropyFlat(axis=1)
        self.chip_size = chip_size
        self.inverse_class_mapping = {}
        self.index_dir = index_dir
        for k, v in self.class_mapping.items():
            self.inverse_class_mapping[v] = k
        if is_no_color(list(color_mapping.values())):
            self.cmap = 'tab20'  ## compute cmap from palette
            import matplotlib as mpl
            bounds = list(color_mapping.keys())
            if len(bounds) < 3: # Two handle two classes i am adding one number to the classes which is not already in bounds
                bounds = bounds + [max(bounds)+1]
            self.mplnorm = mpl.colors.BoundaryNorm(bounds, len(bounds))
        else:
            import matplotlib as mpl
            bounds = list(color_mapping.keys())
            if len(bounds) < 3: # Two handle two classes i am adding one number to the classes which is not already in bounds
                bounds = bounds + [max(bounds)+1]
            self.cmap = mpl.colors.ListedColormap(np.array(list(color_mapping.values()))/255)
            self.mplnorm = mpl.colors.BoundaryNorm(bounds, self.cmap.N)

        if len(color_mapping.keys()) == 1:
            self.cmap = 'tab20'
            self.mplnorm = None

    def open(self, fn):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning) # EXIF warning from TiffPlugin
            if len(fn) != 0:
                img_shape = ArcGISMSImage.read_image(fn[0]).shape
            else:
                labeled_mask = torch.zeros((len(self.class_mapping), self.chip_size, self.chip_size))
                return ArcGISImageSegment(labeled_mask, cmap=self.cmap, norm=self.mplnorm)
            k = 0

            labeled_mask = np.zeros((1, img_shape[0], img_shape[1]))

            for j in range(len(self.class_mapping)):

                if k < len(fn):
                    lbl_name = int(self.index_dir[self.inverse_class_mapping[fn[k].parent.name]])
                else:
                    lbl_name = len(self.class_mapping) + 2
                if lbl_name == j+1:                    
                    img = ArcGISMSImage.read_image(fn[k])
                    k = k + 1
                    if len(img.shape)==3:
                        img = img.transpose(2,0,1)
                        img_mask = img[0]
                        for i in range(1, img.shape[0]):
                            max_unique = np.max(np.unique(img_mask))
                            img_i = np.where(img[i]>0, img[i] + max_unique, img[i])
                            img_mask += img_i
                        img_mask = np.expand_dims(img_mask, axis = 0)
                    else:
                        img_mask = np.expand_dims(img, axis = 0)
                else:
                    img_mask = np.zeros((1, img_shape[0], img_shape[1]))
                labeled_mask = np.append(labeled_mask, img_mask, axis = 0)
            labeled_mask = labeled_mask[1:,:,:]
            labeled_mask = torch.Tensor(list(labeled_mask))
        return ArcGISImageSegment(labeled_mask, cmap=self.cmap, norm=self.mplnorm)

    def reconstruct(self, t): 
        return ArcGISImageSegment(t, cmap=self.cmap, norm=self.mplnorm)

class ArcGISInstanceSegmentationItemList(ImageList):
    "`ItemList` suitable for segmentation tasks."
    _label_cls, _square_show_res = ArcGISSegmentationLabelList, False
    _div = None
    _imagery_type = None
    def open(self, fn):
        return ArcGISMSImage.open(fn, div=self._div, imagery_type=self._imagery_type)

class ArcGISInstanceSegmentationMSItemList(ArcGISInstanceSegmentationItemList):
    "`ItemList` suitable for segmentation tasks."
    _label_cls, _square_show_res = ArcGISSegmentationLabelList, False
    def open(self, fn):
        return ArcGISMSImage.open_gdal(fn)

def mask_rcnn_loss(loss_value, *args):

    final_loss = 0.
    for i in loss_value.values():
        i[torch.isnan(i)] = 0.
        i[torch.isinf(i)] = 0.
        final_loss += i
        
    return final_loss

def mask_to_dict(last_target, device):
    target_list = []
    
    for i in range(len(last_target)):

        boxes =  []
        masks = np.zeros((1, last_target[i].shape[1], last_target[i].shape[2]))
        labels = []
        for j in range(last_target[i].shape[0]):

            mask = np.array(last_target[i].data[j].cpu())
            obj_ids = np.unique(mask)

            if len(obj_ids)==1:
                continue

            obj_ids = obj_ids[1:]
            mask_j = mask == obj_ids[:, None, None]
            num_objs = len(obj_ids)

            for k in range(num_objs):
                pos = np.where(mask_j[k])
                xmin = np.min(pos[1])
                xmax = np.max(pos[1])
                ymin = np.min(pos[0])
                ymax = np.max(pos[0])
                if xmax-xmin==0:
                    xmax += 1
                if ymax-ymin==0:
                    ymax += 1
                boxes.append([xmin, ymin, xmax, ymax])

            masks = np.append(masks, mask_j, axis = 0)
            labels_j = torch.ones((num_objs,), dtype=torch.int64)
            labels_j = labels_j*(j+1)
            labels.append(labels_j)
        
        if(masks.shape[0]==1): # if no object in image
            masks[0,50:51,50:51] = 1
            labels = torch.tensor([0])
            boxes = torch.tensor([[50.,50.,51.,51.]])
        else:
            labels = torch.cat(labels)
            boxes = torch.as_tensor(boxes, dtype=torch.float32)
            masks = masks[1:,:,:]
        masks = torch.as_tensor(masks, dtype=torch.uint8)
        target = {}
        target["boxes"] = boxes.to(device)
        target["labels"] = labels.to(device)
        target["masks"] = masks.to(device)
        target_list.append(target)

    return target_list

class train_callback(LearnerCallback):

    def __init__(self, learn):
        super().__init__(learn)
   
    def on_batch_begin(self, last_input, last_target, **kwargs):
        "Handle new batch `xb`,`yb` in `train` or validation."      
        target_list = mask_to_dict(last_target, self.c_device)
        self.learn.model.train()
        last_input = [list(last_input), target_list]
        last_target = [torch.tensor([1]) for i in last_target]
        return {'last_input':last_input, 'last_target':last_target}

def masks_iou(masks1, masks2):
    # Mask R-CNN

    # The MIT License (MIT)

    # Copyright (c) 2017 Matterport, Inc.

    # Permission is hereby granted, free of charge, to any person obtaining a copy
    # of this software and associated documentation files (the "Software"), to deal
    # in the Software without restriction, including without limitation the rights
    # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    # copies of the Software, and to permit persons to whom the Software is
    # furnished to do so, subject to the following conditions:

    # The above copyright notice and this permission notice shall be included in
    # all copies or substantial portions of the Software.

    # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    # THE SOFTWARE.

    #Method is based on https://github.com/matterport/Mask_RCNN

    if masks1.shape[0] == 0 or masks2.shape[0] == 0:
        return torch.zeros((masks1.shape[0], masks2.shape[0]))
    masks1 = masks1.permute(1,2,0)
    masks2 = masks2.permute(1,2,0)
    masks1 = torch.reshape(masks1 > .5, (-1, masks1.shape[-1])).type(torch.float64)
    masks2 = torch.reshape(masks2 > .5, (-1, masks2.shape[-1])).type(torch.float64)
    area1 = torch.sum(masks1, dim=0)
    area2 = torch.sum(masks2, dim=0)

    intersections = torch.mm(masks1.transpose(1,0), masks2)
    union = area1[:, None] + area2[None, :] - intersections
    overlaps = intersections / union
    return overlaps

def compute_matches(gt_class_ids, gt_masks,
                    pred_class_ids, pred_scores, pred_masks,
                    iou_threshold=0.5, detect_threshold=0.5):

    #Method is based on https://github.com/matterport/Mask_RCNN
    indices = torch.argsort(pred_scores, descending=True)
    pred_class_ids = pred_class_ids[indices]
    pred_scores = pred_scores[indices]
    pred_masks = pred_masks[indices]

    ious_mask = masks_iou(pred_masks, gt_masks)

    pred_match = -1 * np.ones([pred_masks.shape[0]])
    if 0 not in ious_mask.shape:
        max_iou, matches = ious_mask.max(1)
        detected = []
        for i in range(len(pred_class_ids)):
            if max_iou[i] >= iou_threshold and pred_scores[i] >= detect_threshold and matches[i] not in detected and gt_class_ids[matches[i]] == pred_class_ids[i]:
                detected.append(matches[i])
                pred_match[i] = pred_class_ids[i]

    return pred_match

def compute_ap(gt_class_ids, gt_masks,
               pred_class_ids, pred_scores, pred_masks,
               iou_threshold=0.5, detect_threshold=0.5):

    #Method is based on https://github.com/matterport/Mask_RCNN
    pred_match = compute_matches(
        gt_class_ids, gt_masks,
        pred_class_ids, pred_scores, pred_masks,
        iou_threshold, detect_threshold)

    precisions = np.cumsum(pred_match > -1) / (np.arange(len(pred_match)) + 1)
    recalls = np.cumsum(pred_match > -1).astype(np.float32) / len(gt_class_ids)

    precisions = np.concatenate([[0], precisions, [0]])
    recalls = np.concatenate([[0], recalls, [1]])

    for i in range(len(precisions) - 2, -1, -1):
        precisions[i] = np.maximum(precisions[i], precisions[i + 1])

    indices = np.where(recalls[:-1] != recalls[1:])[0] + 1
    mAP = np.sum((recalls[indices] - recalls[indices - 1]) *
                 precisions[indices])
    return mAP

def compute_class_AP(model, dl, n_classes, show_progress, detect_thresh=0.5, iou_thresh=0.5, mean=False):

    model.learn.model.eval()
    if mean:
        aps = []
    else:
        aps = [[] for _ in range(n_classes)]
    with torch.no_grad():
        for input,target in progress_bar(dl, display=show_progress):
            predictions = model.learn.model(list(input))
            ground_truth = mask_to_dict(target, model._device)
            for i in range(len(predictions)):

                predictions[i]["masks"] = predictions[i]["masks"].squeeze()
                if predictions[i]["masks"].shape[0] == 0:
                    continue
                if len(predictions[i]["masks"].shape) == 2:
                    predictions[i]["masks"] = predictions[i]["masks"][None]
                if mean:
                    ap = compute_ap(ground_truth[i]["labels"],
                                    ground_truth[i]["masks"],
                                    predictions[i]["labels"],
                                    predictions[i]["scores"],
                                    predictions[i]["masks"],
                                    iou_thresh,
                                    detect_thresh)
                    aps.append(ap)
                else:
                    for k in range(1, n_classes+1):
                        gt_labels_index = (ground_truth[i]["labels"]== k).nonzero().reshape(-1)
                        gt_labels = ground_truth[i]["labels"][gt_labels_index]
                        gt_masks = ground_truth[i]["masks"][gt_labels_index]
                        pred_labels_index = (predictions[i]["labels"]== k).nonzero().reshape(-1)
                        pred_labels = predictions[i]["labels"][pred_labels_index]
                        pred_masks = predictions[i]["masks"][pred_labels_index]
                        pred_scores = predictions[i]["scores"][pred_labels_index]
                        if len(gt_labels):
                            ap = compute_ap(gt_labels,
                                            gt_masks,
                                            pred_labels,
                                            pred_scores,
                                            pred_masks,
                                            iou_thresh,
                                            detect_thresh)
                            aps[k-1].append(ap)
    if mean:
        if aps != []:
            aps = np.mean(aps, axis=0)
        else:
            return 0.0
    else:
        for i in range(n_classes):
            if aps[i] != []:
                aps[i] = np.mean(aps[i])
            else:
                aps[i] = 0.0
    if model._device == torch.device('cuda'):
        torch.cuda.empty_cache()
    return aps



#Code from https://github.com/fastai/course-v3/blob/master/nbs/dl2/cyclegan_ws.ipynb & https://github.com/eriklindernoren/PyTorch-GAN
import torch
from fastai.vision import nn, Callable, List, LearnerCallback, optim, ifnone, F, flatten_model, requires_grad, SmoothenValue, add_metrics
from .._utils.cyclegan import calculate_activation_statistics, calculate_frechet_distance
from fastprogress.fastprogress import progress_bar
from .._utils.superres import psnr, ssim

import torch.nn as nn
import torch.nn.functional as F
import torch

def weights_init_normal(m):
    classname = m.__class__.__name__
    if classname.find("Conv") != -1:
        torch.nn.init.normal_(m.weight.data, 0.0, 0.02)
    elif classname.find("BatchNorm2d") != -1:
        torch.nn.init.normal_(m.weight.data, 1.0, 0.02)
        torch.nn.init.constant_(m.bias.data, 0.0)


##############################
#           U-NET
##############################


class UNetDown(nn.Module):
    def __init__(self, in_size, out_size, normalize=True, dropout=0.0):
        super(UNetDown, self).__init__()
        layers = [nn.Conv2d(in_size, out_size, 4, 2, 1, bias=False)]
        if normalize:
            layers.append(nn.InstanceNorm2d(out_size))
        layers.append(nn.LeakyReLU(0.2))
        if dropout:
            layers.append(nn.Dropout(dropout))
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)


class UNetUp(nn.Module):
    def __init__(self, in_size, out_size, dropout=0.0):
        super(UNetUp, self).__init__()
        layers = [
            nn.ConvTranspose2d(in_size, out_size, 4, 2, 1, bias=False),
            nn.InstanceNorm2d(out_size),
            nn.ReLU(inplace=True),
        ]
        if dropout:
            layers.append(nn.Dropout(dropout))

        self.model = nn.Sequential(*layers)

    def forward(self, x, skip_input):
        x = self.model(x)
        x = torch.cat((x, skip_input), 1)

        return x


class GeneratorUNet(nn.Module):
    def __init__(self, in_channels=3, out_channels=3):
        super(GeneratorUNet, self).__init__()

        self.down1 = UNetDown(in_channels, 64, normalize=False)
        self.down2 = UNetDown(64, 128)
        self.down3 = UNetDown(128, 256)
        self.down4 = UNetDown(256, 512, dropout=0.5)
        self.down5 = UNetDown(512, 512, dropout=0.5)
        self.down6 = UNetDown(512, 512, dropout=0.5)
        self.down7 = UNetDown(512, 512, dropout=0.5)
        self.down8 = UNetDown(512, 512, normalize=False, dropout=0.5)

        self.up1 = UNetUp(512, 512, dropout=0.5)
        self.up2 = UNetUp(1024, 512, dropout=0.5)
        self.up3 = UNetUp(1024, 512, dropout=0.5)
        self.up4 = UNetUp(1024, 512, dropout=0.5)
        self.up5 = UNetUp(1024, 256)
        self.up6 = UNetUp(512, 128)
        self.up7 = UNetUp(256, 64)

        self.final = nn.Sequential(
            nn.Upsample(scale_factor=2),
            nn.ZeroPad2d((1, 0, 1, 0)),
            nn.Conv2d(128, out_channels, 4, padding=1),
            nn.Tanh(),
        )

    def forward(self, x):
        # U-Net generator with skip connections from encoder to decoder
        d1 = self.down1(x)
        d2 = self.down2(d1)
        d3 = self.down3(d2)
        d4 = self.down4(d3)
        d5 = self.down5(d4)
        d6 = self.down6(d5)
        d7 = self.down7(d6)
        d8 = self.down8(d7)
        u1 = self.up1(d8, d7)
        u2 = self.up2(u1, d6)
        u3 = self.up3(u2, d5)
        u4 = self.up4(u3, d4)
        u5 = self.up5(u4, d3)
        u6 = self.up6(u5, d2)
        u7 = self.up7(u6, d1)

        return self.final(u7)


##############################
#        Discriminator
##############################


class Discriminator(nn.Module):
    def __init__(self, in_channels=3):
        super(Discriminator, self).__init__()

        def discriminator_block(in_filters, out_filters, normalization=True):
            """Returns downsampling layers of each discriminator block"""
            layers = [nn.Conv2d(in_filters, out_filters, 4, stride=2, padding=1)]
            if normalization:
                layers.append(nn.InstanceNorm2d(out_filters))
            layers.append(nn.LeakyReLU(0.2, inplace=True))
            return layers

        self.model = nn.Sequential(
            *discriminator_block(in_channels * 2, 64, normalization=False),
            *discriminator_block(64, 128),
            *discriminator_block(128, 256),
            *discriminator_block(256, 512),
            nn.ZeroPad2d((1, 0, 1, 0)),
            nn.Conv2d(512, 1, 4, padding=1, bias=False)
        )

    def forward(self, img_A, img_B):
        # Concatenate image and condition image by channels to produce input
        img_input = torch.cat((img_A, img_B), 1)
        return self.model(img_input)

class pix2pix(nn.Module):
    
    def __init__(self, ch_in:int, ch_out:int):
        super().__init__()
        
        self.D = Discriminator(ch_in)
        self.G = GeneratorUNet(ch_in, ch_out)
        self.arcgis_results = False
        
    def forward(self, real_A, real_B):
        fake_B = self.G(real_A)
        if self.arcgis_results: return torch.cat([fake_B[:,None],fake_B[:,None]], 1)
        #if not self.training: 
        return [fake_B]

class AdaptiveLoss(nn.Module):
    def __init__(self, crit):
        super().__init__()
        self.crit = crit
    
    def forward(self, output, target:bool, **kwargs):
        targ = output.new_ones(*output.size()) if target else output.new_zeros(*output.size())
        return self.crit(output, targ, **kwargs)

class Adaptivel1Loss(nn.Module):
    def __init__(self, crit1):
        super().__init__()
        self.crit1 = crit1
    
    def forward(self, output, target:bool, **kwargs):
        targ = output.new_ones(*output.size()) if target else output.new_zeros(*output.size())
        return self.crit1(output, targ, **kwargs)

class pix2pixLoss(nn.Module):
    
    def __init__(self, cgan:nn.Module, lambda_A:float=100., lambda_B:float=100, lambda_idt:float=0.5, lsgan:bool=False):
        super().__init__()
        self.cgan,self.l_A,self.l_B,self.l_idt = cgan,lambda_A,lambda_B,lambda_idt
        self.crit = AdaptiveLoss(F.binary_cross_entropy_with_logits)
        self.crit1 = Adaptivel1Loss(F.mse_loss)
    
    def set_input(self, input):
        self.real_A,self.real_B = input

    def forward(self, output, target):
        fake_B = output[0]
        
        self.gen_loss = self.crit(self.cgan.D(fake_B, self.real_A), True)
        self.l1_loss = torch.mean(F.l1_loss(fake_B, self.real_B))

        return self.gen_loss + (self.l_A * self.l1_loss)

class pix2pixTrainer(LearnerCallback):
    _order = -20 #Need to run before the Recorder
    def _set_trainable(self, D=False):
        gen = (not D)
        requires_grad(self.learn.model.G, gen)
        requires_grad(self.learn.model.D, D)
        if not gen:
            self.opt_D.lr, self.opt_D.mom = self.learn.opt.lr, self.learn.opt.mom
            self.opt_D.wd, self.opt_D.beta = self.learn.opt.wd, self.learn.opt.beta
    
    def on_train_begin(self, **kwargs):
        self.G = self.learn.model.G
        self.D = self.learn.model.D
        self.crit = self.learn.loss_func.crit
        self.crit1 = self.learn.loss_func.crit1

        if not getattr(self,'opt_G',None):
            self.opt_G = self.learn.opt.new([nn.Sequential(*flatten_model(self.G))])
        else: 
            self.opt_G.lr,self.opt_G.wd = self.opt.lr,self.opt.wd
            self.opt_G.mom,self.opt_G.beta = self.opt.mom,self.opt.beta

        if not getattr(self,'opt_D',None):
            self.opt_D = self.learn.opt.new([nn.Sequential(*flatten_model(self.D))])

        self.learn.opt.opt = self.opt_G.opt
        self._set_trainable()
        self.gen_smter,self.l1_smter = SmoothenValue(0.98),SmoothenValue(0.98)
        self.d_smter = SmoothenValue(0.98)
        self.recorder.add_metric_names(['gen_loss', 'l1_loss', 'D_loss'])
        
    def on_batch_begin(self, last_input, **kwargs):
        self.learn.loss_func.set_input(last_input)
    
    def on_backward_begin(self, **kwargs):
        self.l1_smter.add_value(self.loss_func.l1_loss.detach().cpu())
        self.gen_smter.add_value(self.loss_func.gen_loss.detach().cpu())
    
    def on_batch_end(self, last_input, last_output, **kwargs):
        self.G.zero_grad()
        fake_B = last_output[0].detach()
        real_A, real_B = last_input

        self._set_trainable(D=True)

        self.D.zero_grad()

        loss_D = 0.5 * (torch.mean(self.crit1(self.D(real_A, real_B), True)) + torch.mean(self.crit1(self.D(fake_B,real_A), False)))

        self.d_smter.add_value(loss_D.detach().cpu())
        if self.learn.model.training == True:
            loss_D.backward()

        self.opt_D.step()

        self._set_trainable()
        
    def on_epoch_end(self, last_metrics, **kwargs):
        return add_metrics(last_metrics, [s.smooth for s in [self.gen_smter,self.l1_smter,self.d_smter]])

def compute_fid_metric(model, data):
    input_a = []
    input_b = []
    pred_b = []
    for input, target in data.valid_dl:
        input_a.append(input[0]/2+0.5)
        input_b.append(input[1]/2+0.5)
        pred = model.learn.pred_batch(batch=(input,target))
        pred_b.append(pred[0]/2+0.5)
    
    data_len = len(data.valid_ds)
    batch_size = data.batch_size

    m1_b, s1_b = calculate_activation_statistics(batch_size, data_len, input_b)
    m2_b, s2_b = calculate_activation_statistics(batch_size, data_len, pred_b)

    fid_value = calculate_frechet_distance(m1_b, s1_b, m2_b, s2_b)

    return fid_value

def compute_metrics(model, dl, show_progress):
    avg_psnr = 0
    avg_ssim = 0
    model.learn.model.eval()
    with torch.no_grad():
        for input, target in progress_bar(dl, display=False):
            prediction = model.learn.model(input[0], input[1])
            avg_psnr += psnr(prediction[0], input[1])
            avg_ssim += ssim(prediction[0], input[1])
    return avg_psnr/len(dl), avg_ssim.item()/len(dl)


import uuid
import arcgis


class SidePanel(object):
    """
    Creates a dashboard Side Panel widget.

    =========================   ===========================================
    **Argument**                **Description**
    -------------------------   -------------------------------------------
    title                       Optional string. Title of the header.
    -------------------------   -------------------------------------------
    description                 Optional string. Description of the widget.
    =========================   ===========================================
    """
    def __init__(self, title=None, description=None):
        # Side Panel starts here

        self._title = title
        self._description = description
        self.type = "leftPanel"

        self._text_color = None
        self._background_color = None

        self._allow_sliding = False

        self._selectors = []
    
    def _repr_html_(self):
        from arcgis.apps.dashboard import Dashboard
        url = Dashboard._publish_random(self)
        return f"""<iframe src={url} width=300 height=300>"""

    @property
    def title(self):
        """
        :return: Title of the widget.
        """
        return self._title
    
    @title.setter
    def title(self, value):
        """
        Set title of the widget.
        """
        if isinstance(value, str):
            self._title = value
    
    @property
    def description(self):
        """
        :return: Description of the widget.
        """
        return self.description
    
    @description.setter
    def description(self, value):
        """
        Set description of the widget.
        """
        if isinstance(value, str):
            self.description = value

    @property
    def text_color(self):
        """
        :return: Text color of the side panel.
        """
        return self._text_color

    @text_color.setter
    def text_color(self, value):
        """
        :return: Set Text color of the side panel widget. Hex color.
        """
        self._text_color = value
        if not isinstance(value, str):
            self._text_color = '#ffffff'

    @property
    def background_color(self):
        """
        :return: Background color of the side panel widget.
        """
        return self._background_color

    @background_color.setter
    def background_color(self, value):
        """
        Set background color of the side panel widget.
        """
        self._background_color = value
        if not isinstance(value, str):
            self._background_color = '#000000'
    
    @property
    def allow_sliding(self):
        """
        :return: True or False for sliding enabled or disabled.
        """
        return self._allow_sliding

    @allow_sliding.setter
    def allow_sliding(self, value):
        """
        Set True or False to enable sliding or to disable sliding.
        """
        self._allow_sliding = value
        if not isinstance(value, bool):
            self._allow_sliding = False

    def add_selector(self, selector):
        """
        Add Number Selector, Category Selector or Date Picker widget.
        """
        self._selectors.append(selector)

    def _convert_to_json(self):
        side_panel = {
            'type': 'leftPanel',
            'allowSliding': self._allow_sliding,
            'selectors': [selector._convert_to_json() for selector in self._selectors]
        }

        if self._background_color:
            side_panel['backgroundColor'] = self._background_color

        if self._text_color:
            side_panel['textColor'] = self._text_color

        if self._title:
            side_panel['title'] = self._title
        
        if self._description:
            side_panel['description'] = self._description
        
        return side_panel



import PIL
try:
    import torch
    from torchvision.models import vgg16_bn
    import torch.nn.functional as F
    import torch.nn as nn
    from torch.nn import Module as NnModule
    from fastai.vision import resize_to
    from fastai.callbacks import hook_outputs
    from fastai.torch_core import requires_grad, children
    from fastprogress.fastprogress import progress_bar
    from .._utils.superres import psnr, ssim
    HAS_FASTAI = True
except Exception as e:
    #import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    class NnModule():
        pass
    HAS_FASTAI = False


def resize_one(fn, i, path_lr, size, path_hr, img_size):
    dest = path_lr/fn.relative_to(path_hr)
    dest.parent.mkdir(parents=True, exist_ok=True)
    img = PIL.Image.open(fn)
    targ_sz = resize_to(img, size, use_min=True)
    img = img.resize(targ_sz, resample=PIL.Image.BILINEAR).convert('RGB')
    #img = img.resize((img_size,img_size), resample=PIL.Image.BILINEAR).convert('RGB')
    dest = dest.with_suffix(".jpeg")
    img.save(dest)


def gram_matrix(x):
    n,c,h,w = x.size()
    x = x.view(n, c, -1)
    return (x @ x.transpose(1,2))/(c*h*w)

class FeatureLoss(nn.Module):
    def __init__(self, m_feat, layer_ids, layer_wgts, base_loss=F.l1_loss):
        super().__init__()
        self.m_feat = m_feat
        self.base_loss = base_loss
        self.loss_features = [self.m_feat[i] for i in layer_ids]
        self.hooks = hook_outputs(self.loss_features, detach=False)
        self.wgts = layer_wgts
        self.metric_names = ['pixel',]

    def make_features(self, x, clone=False):
        self.m_feat(x)
        return [(o.clone() if clone else o) for o in self.hooks.stored]
    
    def forward(self, input, target):
        out_feat = self.make_features(target, clone=True)
        in_feat = self.make_features(input)
        self.feat_losses = [self.base_loss(input,target)]
        self.feat_losses += [self.base_loss(f_in, f_out)*w
                             for f_in, f_out, w in zip(in_feat, out_feat, self.wgts)]
        self.feat_losses += [self.base_loss(gram_matrix(f_in), gram_matrix(f_out))*w**2 * 5e3
                             for f_in, f_out, w in zip(in_feat, out_feat, self.wgts)]
        self.metrics = dict(zip(self.metric_names, self.feat_losses))
        return sum(self.feat_losses)
    
    def __del__(self): self.hooks.remove()

def create_loss(device_type='cuda'):
    if device_type=='cuda':
        vgg_m = vgg16_bn(True).features.cuda().eval()
    else:
        vgg_m = vgg16_bn(True).features.eval()
    requires_grad(vgg_m, False)
    blocks = [i-1 for i,o in enumerate(children(vgg_m)) if isinstance(o,nn.MaxPool2d)]
    blocks, [vgg_m[i] for i in blocks]
    feat_loss = FeatureLoss(vgg_m, blocks[2:5], [5,15,2])
    return feat_loss

def compute_metrics(model, dl, show_progress):
    avg_psnr = 0
    avg_ssim = 0
    model.learn.model.eval()
    with torch.no_grad():
        for input, target in progress_bar(dl, display=False):
            prediction = model.learn.model(input)
            avg_psnr += psnr(prediction, target)
            avg_ssim += ssim(prediction, target)
    return avg_psnr/len(dl), avg_ssim.item()/len(dl)

def get_resize(y, z, max_size, f):
    if y*f <= max_size and z*f <= max_size:
        y_new = y*f
        z_new = z*f
    else:
        if y > z:
            y_new = max_size
            z_new = int(round_up_to_even(z * max_size / y))
        else:
            z_new = max_size
            y_new = int(round_up_to_even(y * max_size / z))
    return (y_new, z_new)


